#!/usr/bin/env python2

#######################################################
#######################################################
#
# freedbtool.py - a python program for interacting with the freedb.org database
#
# Copyright (C) 2002  Robert Muth
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, version 2 of the License.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
# This software has been successfully tested with python 2.1
#
# For more information, invoke as "freedbtool.py -h"
#
# $Id: freedbtool.py,v 1.11 2004/06/05 17:08:38 muth Exp $
#
#######################################################
#######################################################

import urllib
import urllib2
import string
import socket
import os
import sys
import getopt
import re
import getpass

#######################################################
## 
#######################################################

def Error(message):
    print "ERROR: ",message
    sys.exit(-1)

#######################################################
## 
#######################################################

def Info(message):
    print "INFO: ",message
    return

#######################################################
## 
#######################################################
    
def Warning(message):
    print "WARNING: ",message
    return
#######################################################
## read an id file
#######################################################
    
def read_id(filename):
    input = open(filename, "r")
    lines = input.readlines()
    input.close()

    tracks = string.split(lines[0])
#    print tracks
    return tracks

#######################################################
## read an index file
#######################################################

def read_index(filename):
    input = open(filename, "r")
    lines = input.readlines()
    input.close()

    if len(lines) > 1:
        Error("multiple lines in index\n")

    disc = string.split(lines[0])
    
#    print tracks
    return disc

#######################################################
## read an toc file
#######################################################

def read_toc(filename):
    PatternDiscGenre = re.compile(r'^DGENRE\=(\S+)\s*$')
    PatternDiscId = re.compile(r'^DISCID\=(.*)$')
    discids = None
    genre = None 

    input = open(filename, "r")
    lines = input.readlines()
    input.close()

    for line in lines:
        match = PatternDiscGenre.search(line)
        if match:
            genre  = match.group(1)
        match = PatternDiscId.search(line)
        if match:
            discids  = string.split(match.group(1),",")
            for i in range(len(discids)):
                discids[i] = string.strip(discids[i])
    if genre:
        genre = string.lower(genre)
    return (lines,discids,genre)

#######################################################
## 
#######################################################

def increment_revision(lines):
    PatternRevision = re.compile(r'^# Revision:\s+([0-9]+)\s*$')
    for i in range(len(lines)):
        match = PatternRevision.search(lines[i])
        if match: 
            revision = int(match.group(1))
            Info("found revision %d" % revision)
            lines[i] = "# Revision: %d\r\n" % (revision+1)
            return 
    if revision < 0:
        Error("no revision found")
    return 

#######################################################
## 
#######################################################

def replace_submiter(Config, lines):
    PatternRevision = re.compile(r'^# Submitted via:.*$')
    for i in range(len(lines)):
        match = PatternRevision.search(lines[i])
        if match: 
            lines[i] = "# Submitted via: freedbtool %s\r\n" % Config['version']
            return
    return    

#######################################################
## 
#######################################################

def backup_file(filename):
    for i in range(8,0,-1):
        try:
            src = filename + (".%d" % i)
            dst = filename + (".%d" % (i+1))
            #Info("Renaming %s -> %s" % (src,dst))
            os.rename(src,dst)
        except: pass
    try:
        os.rename(filename,filename + ".1")
    except: pass
    
#######################################################
## 
#######################################################

def write_toc(filename,lines):
    backup_file(filename)
    file = open(filename,"w")
    file.write(string.join(lines,""))
    file.close()

#######################################################
## 
#######################################################

def write_index(filename,line):
    file = open(filename,"w")
    file.write(line)
    file.close()
            
#######################################################
## 
#######################################################

def merge_id_into_toc(Config):
    PatternDiscTrack = re.compile(r'^# Track frame.*$')
    PatternDiscLength = re.compile(r'^# Disc length.*$')

    tracks = read_id(Config['file_id'])
    length = tracks[-1]
    tracks = tracks[2:-1]
    
    (lines,discids,genre) = read_toc(Config['file_toc'])

    
    for  i in range(len(lines)):
        if PatternDiscTrack.search(lines[i]):
           break 

    i = i+1
    for t in range(len(tracks)):
        newline  =  "# " + tracks[t] + "\r\n"
        Info("Replacing >" + lines[i+t][:-1] + "< with >" + newline[:-1] + "<")
        lines[i+t] = newline
        
    for  i in range(len(lines)):
        if PatternDiscLength.search(lines[i]):
           break 

    newline =  "# Disc length: " + length + "\r\n"
    Info("Replacing >" + lines[i][:-1] + "< with >" + newline[:-1] + "<")
    write_toc(Config['file_toc'],lines)

#######################################################
## 
#######################################################

def merge_index_into_toc(Config):
    PatternDiscYear = re.compile(r'^DYEAR\=.*$')
    PatternDiscTitle = re.compile(r'^DTITLE\=.*$')

    index = read_index(Config['file_index'])
    Info("found index: >%s< >%s<" % (index[0],index[1]))
    (lines,discids,genre) = read_toc(Config['file_toc'])
    if discids.count(index[1]) == 0:
        Error("id missmatch: >"+ string.join(discids,",")+"<")
    if genre:
        Error("genre already defined: >" + genre + "<")
    for  i in range(len(lines)):
        if not PatternDiscTitle.search(lines[i]): continue
        if PatternDiscYear.search(lines[i+1]):
            Info("updating genre " + index[0] )
            lines[i+2] = "DGENRE=" + index[0] + "\r\n"
        else:
            Info("adding genre " + index[0] )
            lines.insert(i+1,"DGENRE=" + index[0] + "\r\n")
            lines.insert(i+1,"DYEAR=\r\n")
    write_toc(Config['file_toc'],lines)

#######################################################
## 
#######################################################

def make_command_url(Config,command):
    return "%s%s?cmd=%s&hello=%s+%s+%s+%s&proto=%s" % \
           (Config['server'],Config['cgi_query'],command,Config['user'], Config['host'], Config['client'], Config['version'], Config['protocol'])

#######################################################
## 
#######################################################

def cddb_motd(Config):

    url = make_command_url(Config, "motd")

    response = urllib2.urlopen(url)

    line = ""
    for x in response.readlines():
        line = line + x

    print line

#######################################################
## 
#######################################################

def cddb_lscat(Config):

    url = make_command_url(Config, "cddb+lscat")

    response = urllib2.urlopen(url)

    line = ""
    for x in response.readlines():
        line = line + x

    print line

#######################################################
## 
#######################################################    

def cddb_query(Config, tracks):

    result = []
    query_str = string.join(tracks)

    #print query_str

    query_str = urllib.quote_plus(string.rstrip(query_str))

    url = make_command_url(Config,"cddb+query+" + query_str  )

    response = urllib2.urlopen(url)

    line = response.readline()
    components = string.split(line)

    code = string.atoi(components[0])
    line = string.join(components[1:])

    if code == 200:
        result.append(line)
        
    elif code == 202:
        pass

    elif code == 211 or code == 210:
        for line in response.readlines():
            if line[0] == '.': break
            result.append(line[0:-1])
    else:
        Error("unknown return code %s\n" % components[0])
        
    return result

#######################################################
## 
#######################################################

def cddb_read(Config,genre,discid):

    url = make_command_url(Config,"cddb+read+" + genre + "+" + discid )

    response = urllib2.urlopen(url)
    
    header = string.split(string.rstrip(response.readline()), ' ', 3)
    lines = response.readlines()
    if lines == []:
        Error("no results returned from database\n")

    line = lines[-1]
    if line[0] != '.':
        Error("toc not terminated")

    return lines[0:-1]
    
#######################################################
## 
#######################################################

def cddb_submit(Config,toc,discid,genre):

    header = {}
    header['Category']  = genre
    header['Discid'] = discid
    header['User-Email'] = Config['user'] + "@" + Config['host'] 
    header['Submit-Mode'] = Config['mode']
    header['Charset'] = 'ISO-8859-1'
    header['X-Cddbd-Note'] = 'Sent by ' + Config['client'] + " " + Config['version']

    url = Config['server'] + Config['cgi_submit']
    #print url,header
    request = urllib2.Request(url,None,header)

    data = string.join(toc,"")
    #data = data[0:-3]
        
    response = urllib2.urlopen(request,data)
 
    for x in response.readlines():
        print x

#######################################################
## 
#######################################################

def version():
    
    (l,v,x) = string.split('$Revision: 1.11 $')
    return v
    
#######################################################
## 
#######################################################

def usage():
    print\
          "Freedbtool Version " + version() + "\n" + \
          "\n" + \
          "Usage: freedbtool [mode] [option]*\n" + \
          "\n" + \
          "Retrieve or submit information to and from the freedb.org database.\n"+ \
          "The mode parameter specifies the concrete command.\n"+ \
          "Options are used to change the default behavior of commands.\n"+ \
          "Currently available modes are: config,lscat,get,merge,submit\n" + \
          "The model of operation of freedb tool is centered around 3 files characterizing a CD\n" + \
          "\n"+ \
          "id\n" + \
          "\t contains track length information for a CD,it is compatible with the cddb query command.\n" + \
          "\t A seperate tool discid is used to obtain this information.\n" + \
          "index\n" + \
          "\t contains genre and checksum information for a CD, it is compatible with the freedb read command.\n" + \
          "\t This information is obtained by quering the freedb.org with the information in id.\n" + \
          "toc\n" + \
          "\t contains the actual freedb.org database record for a CD.\n" + \
          "\t This information is obtained by quering the freedb.org with the information in index.\n" + \
          "\n" + \
          "Options:\n" + \
          "\n" + \
          "-t filename\n" + \
          "\tchange default toc filename [toc]\n" + \
          "-x filename\n" + \
          "\tchange default index filename [index]\n" + \
          "-i filename\n" + \
          "\tchange default id filename [id]\n" + \
          "-n number\n" + \
          "\tspecify record number in case of multiple matches\n" + \
          "-c clientname\n" + \
          "\tclientname used to when contacting the server [freedbtool]\n" + \
          "-h domain\n" + \
          "\tspecify domain part of user email\n" + \
          "-u username\n" + \
          "\tspecify username part of user email\n" + \
          "-s server\n" + \
          "\tspecify server [http://freedb.freedb.org/]\n" + \
          "-d discid-command\n" + \
          "\tspecify discid commandline\n" + \
          "-m db-submit-mode\n" + \
          "\tspecify mode for submit command [submit]\n" + \
          "-p protocolnumber\n" + \
          "\tspecify protocol number [5]\n" + \
          "\n" + \
          "Command description:\n" + \
          "\n"+ \
          "config\n"+ \
          "\tprint version and current defaults\n"+ \
          "motd\n"+ \
          "\tprint message of the day from freedb.org \n"+ \
          "get\n"+ \
          "\tget id, index, toc for current disc from freedb.org \n"+ \
          "update\n"+ \
          "\tupdate existing toc from freedb.org \n"+ \
          "submit\n"+ \
          "\tsubmit revised toc to freedb.org \n"+ \
          "merge\n"+ \
          "\tmerge genre info from index into toc file \n"+ \
          "track\n"+ \
          "\tmerge/update track info from id file into toc file \n"+ \
          "\n" + \
          "Written by Robert Muth (robert at muth dot org)\n" + \
          "See program file for copyright information\n"
    return -1

#######################################################
## 
#######################################################

if __name__ == "__main__": 

    #
    # DEFAULTS
    #
    
    Config = {}
    Config['version'] = version()
    Config['server'] = 'http://freedb.freedb.org/'
    Config['protocol'] = '5'
    Config['user'] = getpass.getuser()
    Config['host'] = socket.gethostname()
    Config['client'] = 'freedbtool'
    Config['mode'] = 'submit'
    Config['cgi_query'] = '~cddb/cddb.cgi'
    Config['cgi_submit'] = '~cddb/submit.cgi'
    Config['discid'] = 'discid /dev/cdrom'
    Config['file_id'] = 'id'
    Config['file_index'] = 'index'
    Config['file_toc'] = 'toc'
    Config['select'] = -1

    #
    # COMMAND LINE PARSING
    #
    
    try:
        mode = sys.argv[1]
        opts, args = getopt.getopt(sys.argv[2:], "m:d:t:n:h:u:x:s:t:i:p:c:")
        for o, a in opts:
            if o == '-n':
                Config['select'] =  int(a)
                continue
            if o == '-t':
                Config['file_toc'] =  a
                continue
            if o == '-x':
                Config['file_index'] =  a
                continue
            if o == '-i':
                Config['file_id'] =  a
                continue
            if o == '-h':
                Config['host'] =  a
                continue
            if o == '-u':
                Config['user'] =  a
                continue
            if o == '-s':
                Config['server'] =  a
                continue
            if o == '-d':
                Config['discid'] =  a
                continue                
            if o == '-m':
                Config['mode'] =  a
                continue
            if o == '-p':
                Config['protocol'] =  a
                continue
            if o == '-c':
                Config['client'] =  a
                continue
                
    except:
        sys.exit( usage() )

    #
    # MODE DISPATCHING
    #
    
    if mode == 'merge':
        merge_index_into_toc(Config) 
    elif mode == 'track':
        merge_id_into_toc(Config) 
    elif mode == 'motd':
        cddb_motd(Config) 
    elif mode == 'lscat':
        cddb_lscat(Config) 
    elif mode == 'config':
        print "Version $Revision: 1.11 $"
        for k in Config.keys():
            print k,Config[k]
    elif mode == 'submit':
        (lines,discids,genre) = read_toc(Config['file_toc'])
        if genre and discids:
            increment_revision(lines)
            replace_submiter(Config,lines)
            Info("submitting: disk " + discids[0] + " " + genre + " as " + Config['user'] + "@" + Config['host'])
            cddb_submit(Config,lines,discids[0],genre)
        else:
            print "missind id/genre"

    elif mode == 'update':
        (lines,discids,genre) = read_toc(Config['file_toc'])
        if not genre:
            Error("toc files does not contain valid DGENRE tag\n")
        genre =string.lower(genre)
        Info("updating %s %s" % (genre,discids[0]))            
        toc = cddb_read(Config,genre,discids[0])
        write_toc(Config['file_toc'],toc)

    elif mode == 'get':
        command =  Config['discid'] + '>' + Config['file_id']
#        print command
        
        os.system( command)
        tracks = read_id( Config['file_id'] )
        results = cddb_query(Config,tracks)
        Info("found %d records" % len(results))
        num_selected = Config['select']
        num_results =  len(results)
        if num_results == 1:
            num_selected = 0

        if 0 <= num_selected and  num_selected < num_results:
            index = results[num_selected]
            Info("using number %d: %s" % (num_selected,index))
            write_index(Config['file_index'],index)
            part = string.split(index)
            toc = cddb_read(Config,part[0],part[1])
            write_toc(Config['file_toc'],toc)
        elif num_results == 0:
            Warning("no matches found.")
        else:
            Warning("multiple matches found. Use -n options to choose the right one.")
            for i in range(num_results):
                index = results[i]
                print "####################################################################################"
                print i,index
                print "####################################################################################"
                part = string.split(index)
                toc = cddb_read(Config,part[0],part[1])
                print string.join(toc)
    else:
        sys.exit( usage() )
    
#######################################################
## 
#######################################################



